#ifndef __resource_pool_header_included_p
#define __resource_pool_header_included_p
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/thread.h>
#include <map>
#include <set>

///@file
///  This file defines template, ResourcePool&lt;Objects,Grade&rt, which
///  allows you to define graded pools of objects of any given object
///  type and grading scheme.  For example, you could create a pool
///  of data base connections and grade them according to throughput
///  limits, etc.  
///
///  The nature of a grade must be encapsulated in Grade -- which must be
///  sortable and copiable.  The Objects themselves do not have to be copiable 
///  because the ResourcePool actually keeps a collection of graded pointers to 
///  objects.
///
///  A ResourcePool has a fixed number of objects of each type and grade
///  provides a mechanism for allocating and freeing objects from and to
///  the pool.
///
///  @see ResourcePool.
///

namespace cxxtls
{

template<class Object, class Grade>
class ResourcePool;

//=====================================================================================
//=====================================================================================
//=====================================================================================

template<class Object, class Grade>
class ScopedBorrow
    ///  A ScopedBorrow is a wrapper around a pointer to an Object which has
    ///  been borrowed from a ResourcePool.  The pointer may be null, meaning
    ///  that no objects of interest were available for borrowing from the
    ///  resource pool.
    ///
    ///  The constructor for ScopedBorrow requires that you pass a resource
    ///  pool to it.  The destructor returns the borrowed pointer back to
    ///  its original owner.
    ///
    ///  An object of this type can be treated like a pointer to an Object.
    ///  To get the actual pointer, as when you need to pass it to function,
    ///  call the get() method.
    ///
    ///  Note that ScopedBorrow objects are not copiable or assignable.
{
    typedef ResourcePool<Object,Grade> Owner;

    Owner   &owner_;
    Grade    grade_;
    Object  *object_;

public:

    ScopedBorrow(Owner &o, Grade const &g, bool up=false, bool dn=false);
        ///< 
        /// Borrow an object of a specified grade from the specified resourcePool.
        /// A null may be the result of the borrowing, so invoke operator bool() to
        /// see if a non-null value was obtained.
        ///
        /// @param o  The resource pool owning the object to be borrowed.
        /// @param g  The desired grade of object to be borrowed.
        /// @param up The flag indicating that an upgrade should be attempted if the
        ///           desired grade is unavailable.   Defaults to false.
        /// @param dn The flag indicating that a downgrade should be attempted if the
        ///           desired grade is unavailable.  Defaults to false.
        ///
    
   ~ScopedBorrow();
       ///< return the borrowed object, if any

   operator bool() const { return object_ != 0; }
       ///<  return true if this object is usable -- ie not null.

   Object& operator*() { return *object_; }
       ///<  return a reference to the object referred to by this handle.

   Object const & operator*() const { return *object_; }
       ///<  return a const reference to the object referred to by this handle.

   Object* operator->() { return object_; }
       ///<  Operator->() for this handle

   Object const* operator->() const { return object_; }
       ///<  const Operator->() for this handle

   Object *get() { return object_; }
       ///<  Get the raw pointer so you can pass it to a function.  Do not store
       ///   the raw pointer value!

   Object const *get() const { return object_; }
       ///<  Get the raw pointer so you can pass it to a function.  Do not store
       ///   the raw pointer value!

private:
    ScopedBorrow();                         // not implemented
    ScopedBorrow(ScopedBorrow const &);     // not implemented
    void operator=(ScopedBorrow const &);   // not implemented
};

//=====================================================================================
//=====================================================================================
//=====================================================================================


template<class Object, class Grade>
class ResourcePool
    /// A collection of graded Object pointers which can be loaned out and
    /// which must be returned.
    ///
    /// This class is thread safe, so you can have a program singleton resource
    /// pool which can be accessed by many threads.
    ///
    /// Pools are empty unless the calling application uses the adopt() method to
    /// insert objects into the pool.  This should be done as early as possible in
    /// the run of an application.
    ///
    /// Only one thread can have a given object of a given type.  To borrow an
    /// object from the resource pool, declare a variable of type ScopedBorrow,
    /// see above, and construct it given the resource pool of interest.  For example:
    ///
    ///@code
    ///
    ///   extern ResourcePool<ObjectType,int> resourcePool;
    ///
    ///   void f()
    ///   {
    ///      ScopedBorrow<ObjectType,int>  handle(resource,14);
    ///
    ///      if(handle)
    ///        handle->memberFunction();
    ///   }
    ///
    ///@endcode
    ///
    ///  In the above example, the handle's destructor returns the
    ///  borrowed object to the resource pool at the end of function, f().
    ///  the parameter "14" is the desired "grade" of pool.  The handle will only
    ///  be non-null if an object of grade 14 can actually be found, since the
    ///  up and dn parameters to the constructur are left in the default state,
    ///  which is "false".
    ///
{
    friend class ScopedBorrow<Object,Grade>;

    typedef std::multimap<Grade, Object*> Objects;

public:

   ResourcePool(){}                        ///<  construct an empty pool
  ~ResourcePool();                         ///<  delete 

   void adopt(Object* o, Grade const & g); ///<  Take ownership of an object, with a specific
                                           ///   grade and delete it upon destruction.  Make
                                           ///   it available for borrowing.
                                           ///
                                           ///   @param o  The object to be added to the pool
                                           ///   @param g  The grade assigned to it.

private:

    Mutex   mutex_;                        ///<  Mutex guarding all data of interest
    Objects available_;                    ///<  unused objects available for allocation

    struct BorrowResult
    {
       Object *rv_; 
       Grade   grade_; // actual grade, not requested.
    };

    BorrowResult allocate(Grade const &g,
                          bool         up,
                          bool         dn
                         );                ///<  Allocate one of the available objects.
                                           ///   Try to find one of the desired grade.
                                           ///   If up an dn are specified, try finding
                                           ///   a match of a higher or lower grade.
                                           ///
                                           ///   @param g  The desired grade of object.
                                           ///   @param up The flag indicating that upgrade
                                           ///             is desired if necessary.
                                           ///   @param dn The flag indicating that downgrade
                                           ///             is desired if necessary.
                                           ///
                                           ///   Upgrades and downgrades only occur if the
                                           ///   desired grade of object cannot be found.

    void free(Object *o, Grade const &g);  ///<  Return a borrowed object to the available_
                                           ///   pool.
    
};

//=====================================================================================
//=====================================================================================
//=====================================================================================

template<class Object, class Grade>
ScopedBorrow<Object,Grade>::ScopedBorrow(Owner &o, Grade const &g, bool up, bool dn)
:  owner_(o)
{
    typedef typename ResourcePool<Object, Grade>::BorrowResult  BorrowResult;

    BorrowResult rv = o.allocate(g, up, dn);

    if(rv.rv_)
    {
       grade_  = rv.grade_;
       object_ = rv.rv_;
    }
    else
    {
      grade_  = g;
      object_ = 0;
    }

}

//=====================================================================================
//=====================================================================================
//=====================================================================================

template<class Object, class Grade>
ScopedBorrow<Object,Grade>::
~ScopedBorrow()
     /// return the borrowed object, if any
{
   if(object_)
   {
      owner_.free(object_, grade_);
   }
}

//=====================================================================================
//=====================================================================================
//=====================================================================================
template<class Object, class Grade>
void 
ResourcePool<Object,Grade>::
free(Object *o, Grade const &g)
{
    // this function can only be called by ScopedBorrow<>

    ScopedMutexLock  lock(mutex_);

    available_.insert( typename Objects::value_type(g,o) );
}

//=====================================================================================
//=====================================================================================
//=====================================================================================
template<class Object, class Grade>
typename ResourcePool<Object,Grade>::BorrowResult 
ResourcePool<Object,Grade>::
allocate(Grade const &g, bool up, bool dn)
{
    ScopedMutexLock  lock(mutex_);

    // This function can only be called from ScopedBorrow<>

    static int debug=false;

    if(debug)
    {
       typename Objects::iterator f = available_.begin(),
                                  l = available_.end();
       while(f != l)
       {
         Grade gtmp = f->first;
         gtmp = gtmp;
         ++f;
       }
    }


    BorrowResult rv;
    
    rv.grade_ = g;
    rv.rv_    = 0;

    typename Objects::iterator where = available_.upper_bound(g);

    // where points to the item just above the desired range

    if(where == available_.begin())
    {
        // no free objects were found with grade <= the desired grade
        // so upgrade is the only choice

        if(!up)
           return rv; // return fail
    }
    else
    {
       // where points to higher grade than desired and there are grades
       // <= to the desired grade

       --where;  // step down to desired grade or less

       if(where->first != g)
       {
          // no objects equal to desired grade are found, so *where must be less than desired

          if(!dn)
          {
              // downgrade is not desired

              if(!up)
                 return rv; // fail because we have no matches and do not want
              else
              {
                 ++where;   // upgrade is ok, so lets step back up to the higher grade

              }
          }
          else
          {
             // downgrade is ok, so lets just return *where
          }
       }
       else
       {
          // where points to an object of the desired grade
       }
    }

    // return whatever we have left
    if(where == available_.end())
    {
        return rv; // no items available
    }


    rv.rv_    = where->second;
    rv.grade_ = where->first;     // return the actual grade

    available_.erase(where);  // remove the allocated object from the pool.

    return rv;
}

//=====================================================================================
//=====================================================================================
//=====================================================================================
template<class Object, class Grade>
void
ResourcePool<Object,Grade>::
adopt(Object*o, Grade const &g)
{
    ScopedMutexLock  lock(mutex_);

    available_.insert( typename Objects::value_type(g,o) );

}

//=====================================================================================
//=====================================================================================
//=====================================================================================
template<class Object, class Grade>
ResourcePool<Object,Grade>::
~ResourcePool()
{
    ScopedMutexLock  lock(mutex_);

    typename Objects::iterator f = available_.begin(),
                               l = available_.end();

    while(f != l)
    {
        typename Objects::value_type &v = *f++;

        delete v.second;
    }

    available_.clear();

}




} // namemspace cxxtls
#endif
